#!/usr/bin/ruby
# coding: utf-8

#######################################
#
# Marcin Woźniak
# s434812
#
########################################

load '../module.rb'
require 'thread'
require 'thwait'

def returnRownanie(a,b,p)
  puts
  puts "Równanie krzywej jest równe: " + "Y^2 = X^3+" + a.inspect + "X+" + b.inspect + " mod "+ p.inspect
  puts
end

def delta(a,b,p)
  d = ((4 * betterExponentiation(a,3,p) % p) + (27 * betterExponentiation(b,2,p) % p)) % p
  return d
end

def rownanieKrzywej(a,b,p,x)
  fx = ((betterExponentiation(x,3,p) + (a * x) % p + b % p) % p) % p
  return fx
end

def generatorKrzywej(p)
  a = 0
  b = 0

  while true
    if primalityTest(p) &&  (p % 4 == 3)
      threads = []

      threads << Thread.new {
        a = SecureRandom.random_number(1..p-1)
      }

      threads << Thread.new {
        b = SecureRandom.random_number(1..p-1)
      }

      ThreadsWait.all_waits(*threads)
      if delta(a,b,p) != 0
        returnRownanie(a,b,p)
        return a,b
      end
    else
      puts "Liczba nie spełnia wymagań"
      break
    end
  end
end

def punktNaKrzywej(a,b,p)
  if (delta(a,b,p) != 0) &&  (p % 4 == 3)
    while true
      x = SecureRandom.random_number(0..p-1)
      fx = rownanieKrzywej(a,b,p,x)
      if remSqEuler(fx,p)
        y = betterExponentiation(fx,((p+1)/4),p)
        return x,y
      end
    end
  end
end

def czyPunktNalezyDoKrzywej(a,b,p,x,y)
  fx = rownanieKrzywej(a,b,p,x)

  if fx == betterExponentiation(y,2,p)
    return true
  else
    return false
  end
end

def punktPrzeciwny(x,y)
  return x,-y
end

def sumaPunktow(a,b,p,x1,y1,x2,y2)
  # 0 - element neutrany --> P + 0 = P
  if (x1 == "e" && y1 == "e" )
    return x2,y2
  elsif (x2 == "e" && y2 == "e")
    return x1,y1
  end
  # P + Q = R
  if (x1 != x2)
    lambda = (((y2 - y1) % p) * reciprocal_Phi_p((x2 - x1),p)) % p
    x3 = (betterExponentiation(lambda,2,p) - (x1 % p) - (x2 % p)) % p
    y3 = (lambda * (x1 - x3) - y1) % p
    return x3,y3
  end
  # P + -Q = 0 DZIALA
  if (x1 == x2) && (y1 == -y2)
    puts "0 - el.neutralny"
    return e,e
  end
  # P + P = 2P DZIALA
  if (x1 == x2) && (y1 == y2)
    lambda = (((3 * betterExponentiation(x1,2,p) % p + a) % p) * reciprocal_Phi_p(2 * y1,p)) % p
    x3 = (betterExponentiation(lambda,2,p) - (x1 % p) - (x2 % p)) % p
    y3 = (lambda * (x1 - x3) - y1) % p
    return x3,y3
  end
end

#puts generatorKrzywej(11).inspect
#puts punktNaKrzywej(2,7,11).inspect
#puts czyPunktNalezyDoKrzywej(2,7,11,7,1).inspect
#puts czyPunktNalezyDoKrzywej(2,7,11,10,10).inspect
#puts punktPrzeciwny(2,2).inspect
#puts sumaPunktow(2,2,17,5,1,5,1).inspect

##################################################
# Zadania z konca wykladu suma punktow
#
# Niech E/F7 będzie postaci
# E : Y^2 = X^3 + 1
#
# Niech P=(1,3)  Q=(2,4)  R=(6,0)
#
# Obliczyc:
# 1) -Q
# 2) R + [0,0]
# 3) P + Q
# 4) 2R
#
# Rozwiązanie:
a = 0
b = 1
p = 7

xp = 1
yp = 3

xq = 2
yq = 4

xr = 6
yr = 0

# 1) -Q
puts punktPrzeciwny(xq,yq).inspect

# 2) R + [0,0]
puts sumaPunktow(a,b,p,xr,yr,"e","e").inspect

# 3) P + Q
puts sumaPunktow(a,b,p,xp,yp,xq,yq).inspect

# 4) 2R
puts sumaPunktow(a,b,p,xp,yp,xp,yp).inspect
##################################################

##################################################
# Zadanie od profersora
a=239614427021073265587611886177902927263167863041565491257781227550405368115731464059190159
b=447169285435982716467332439542997876345372330045685811964291613238129105735899852114277221
p=1183779584357076950937981497685946292711107412152534481102525547387604378262522402526266939
px=285113634279465403319996581740169338329454608669814309137990174814243655992779447106132850
py=598700530906084162596261101440667782569915319623798143751082061599951188013331503150304328

# Zadanie 1
puts
puts "Zadanie 1"
puts generatorKrzywej(p).inspect

# Zadanie 2
puts
puts "Zadanie 2"
puts punktNaKrzywej(a,b,p).inspect

# Zadanie 3
puts
puts "Zadanie 3"
puts czyPunktNalezyDoKrzywej(a,b,p,px,py).inspect

# Zadanie 4
puts
puts "Zadanie 4"
puts punktPrzeciwny(px,py).inspect

# Zadanie 5
puts
puts "Zadanie 5"
q=sumaPunktow(a,b,p,px,py,px,py)
puts "P+P=Q=" + q.inspect

qx=q[0].to_i
qy=q[1].to_i
r=sumaPunktow(a,b,p,px,py,qx,qy)
puts "P+Q=R=" + r.inspect

##################################################
